{-# OPTIONS_GHC -fglasgow-exts #-}

module Edg where {

import List;
import Nat;
import Substitution;
import Term;
import HOL;
import Util;
import Option;

data Hterm a b = HHole | HFun a [(Hterm a b)];

mergeHTerms ::
  forall a b. (Eq a) => [(Hterm a b)] -> [(Hterm a b)] -> Maybe [(Hterm a b)];
mergeHTerms [] [] = Just [];
mergeHTerms [] (uu : uv) = Nothing;
mergeHTerms (uw : ux) [] = Nothing;
mergeHTerms (HHole : hsl) (h : hsr) =
  (case mergeHTerms hsl hsr of {
    Nothing -> Nothing;
    Just res -> Just (h : res);
  });
mergeHTerms (HFun f hs : hsl) (HHole : hsr) =
  (case mergeHTerms hsl hsr of {
    Nothing -> Nothing;
    Just res -> Just (HFun f hs : res);
  });
mergeHTerms (HFun f fs : hsl) (HFun g gs : hsr) =
  (if f == g
    then (case mergeHTerms fs gs of {
           Nothing -> Nothing;
           Just mhs ->
             (case mergeHTerms hsl hsr of {
               Nothing -> Nothing;
               Just res -> Just (HFun f mhs : res);
             });
         })
    else Nothing);

merge :: forall a b. (Eq a) => Hterm a b -> Hterm a b -> Maybe (Hterm a b);
merge hl hr =
  (case mergeHTerms [hl] [hr] of {
    Nothing -> Nothing;
    Just res -> Just (nth res Zero_nat);
  });

mergeX ::
  forall a b.
    (Eq a,
      Eq b) => a -> Hterm b a -> [(Hterm b a, a)] -> Maybe [(Hterm b a, a)];
mergeX x h [] = Just [];
mergeX x h ((ht, y) : ps) =
  (if x == y
    then (case merge h ht of {
           Nothing -> Nothing;
           Just m -> mergeX x m ps;
         })
    else (case mergeX x h ps of {
           Nothing -> Nothing;
           Just pairs -> Just ((ht, y) : pairs);
         }));

mergeAll :: forall a b. (Eq a, Eq b) => [(Hterm a b, b)] -> Bool;
mergeAll [] = True;
mergeAll ((h, x) : ps) =
  (case mergeX x h ps of {
    Nothing -> False;
    Just a -> mergeAll a;
  });

gcMatchList ::
  forall a b. (Eq a) => [(Hterm a b, Term a b)] -> Maybe [(Hterm a b, b)];
gcMatchList [] = Just [];
gcMatchList ((HHole, uu) : ps) = gcMatchList ps;
gcMatchList ((HFun f hs, Fun g ss) : ps) =
  (if f == g
    then (case decomp hs ss of {
           Nothing -> Nothing;
           Just pairs -> gcMatchList (append pairs ps);
         })
    else Nothing);
gcMatchList ((HFun f hs, Var x) : ps) =
  (case gcMatchList ps of {
    Nothing -> Nothing;
    Just pairs -> Just ((HFun f hs, x) : pairs);
  });

gcMatchI :: forall a b. (Eq a, Eq b) => Hterm a b -> Term a b -> Bool;
gcMatchI h t =
  (case gcMatchList [(h, t)] of {
    Nothing -> False;
    Just a -> mergeAll a;
  });

capI ::
  forall a b. (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> Hterm a b;
capI uu (Var uv) = HHole;
capI r (Fun f ts) =
  let {
    h = HFun f (mapa (capI r) ts);
  } in (if testPos (\ ra -> gcMatchI h (fst ra)) r then HHole else h);

capRM2 ::
  forall a b.
    (Eq a, Eq b) => (a -> [(Term a b, Term a b)]) -> Term a b -> Hterm a b;
capRM2 uu (Var uv) = HHole;
capRM2 rm (Fun f ts) =
  let {
    h = HFun f (mapa (capRM2 rm) ts);
  } in (if testPos (\ r -> gcMatchI h (fst r)) (rm f) then HHole else h);

capRM ::
  forall a b.
    (Eq a,
      Eq b) => Maybe (a -> [(Term a b, Term a b)]) -> Term a b -> Hterm a b;
capRM r = (case r of {
            Nothing -> (\ t -> HHole);
            Just a -> capRM2 a;
          });

checkEdgRM ::
  forall a b.
    (Eq a,
      Eq b) => Maybe (a -> [(Term a b, Term a b)]) ->
                 Maybe (a -> [(Term a b, Term a b)]) ->
                   ((Term a b, Term a b), (Term a b, Term a b)) -> Bool;
checkEdgRM rm rrm ((s, t), (u, v)) =
  gcMatchI (capRM rm t) u && gcMatchI (capRM rrm u) t;

reverse_trs :: forall f v. [(Term f v, Term f v)] -> [(Term f v, Term f v)];
reverse_trs r = mapa (\ lr -> (snd lr, fst lr)) r;

}
