{-# OPTIONS_GHC -fglasgow-exts #-}

module ErrorMonad where {

import Util;
import List;
import Nat;
import Product_Type;

returna :: forall a e. a -> Eithera e a;
returna a = Righta a;

errora :: forall e a. e -> Eithera e a;
errora a = Lefta a;

checkB :: forall e. Bool -> e -> Eithera e ();
checkB b e = (if b then returna () else errora e);

isValue :: forall e a. Eithera e a -> Bool;
isValue p = (case p of {
              Lefta l -> False;
              Righta r -> True;
            });

bind :: forall e a b. Eithera e a -> (a -> Eithera e b) -> Eithera e b;
bind p vq = (case p of {
              Lefta a -> Lefta a;
              Righta a -> vq a;
            });

catcha :: forall e a f. Eithera e a -> (e -> Eithera f a) -> Eithera f a;
catcha p eq = (case p of {
                Lefta a -> eq a;
                Righta a -> Righta a;
              });

foldR :: forall a b. (a -> [a] -> b -> b) -> [a] -> b -> b;
foldR f [] b = b;
foldR f (x : xs) b = f x xs (foldR f xs b);

checkF :: forall a b. (a -> Bool) -> (a -> b) -> a -> Eithera b ();
checkF f g x = (if f x then returna () else errora (g x));

checkAllMI :: forall a e. (a -> Eithera e ()) -> [a] -> Eithera (a, e) ();
checkAllMI f xs =
  foldra (\ x iq -> bind (catcha (f x) (\ xa -> errora (x, xa))) (\ xa -> iq))
    xs (returna ());

checkSubset :: forall a. (Eq a) => [a] -> [a] -> Eithera a ();
checkSubset xs ys =
  catcha (checkAllMI (checkF (\ x -> member x ys) id) xs)
    (\ x -> errora (snd x));

checkDisjoint :: forall a. (Eq a) => [a] -> [a] -> Eithera a ();
checkDisjoint xs ys =
  catcha (checkAllMI (checkF (\ x -> not (member x ys)) id) xs)
    (\ x -> errora (snd x));

checkAllCombinations ::
  forall a b. (a -> a -> Eithera b ()) -> [a] -> Eithera ((a, a), b) ();
checkAllCombinations c xs =
  foldR (\ x ys res ->
          catcha
            (checkAllMI
              (\ y ->
                bind (catcha (c x y) (\ xa -> errora ((x, y), xa)))
                  (\ xa -> res))
              ys)
            (\ xa -> errora (snd xa)))
    xs (returna ());

checkPartition :: forall a. (Eq a) => [[a]] -> Eithera a ();
checkPartition cs =
  catcha (checkAllCombinations checkDisjoint cs) (\ x -> errora (snd x));

checkAllMIndex ::
  forall a e.
    (Nat -> a -> Nat) -> (Nat -> a -> Eithera e ()) -> [a] -> Eithera e ();
checkAllMIndex g f xs =
  snd (foldla (\ (a @ (n, mon)) x -> (g n x, bind mon (\ xa -> f n x)))
        (one_nat, returna ()) xs);

}
