{-# OPTIONS_GHC -fglasgow-exts #-}

module SubtermCriterionImpl where {

import Term;
import Nat;
import List;
import Show;
import ErrorMonad;
import DpTransImpl;
import SubtermCriterion;
import Util;
import Product_Type;
import Option;
import HOL;

newtype SpL a = SimpProj [(a, Nat)];

create_sp_representation :: forall a. (Sshow a) => SpL a -> [[Char]];
create_sp_representation (SimpProj []) = [];
create_sp_representation (SimpProj ((f, n) : sp)) =
  concata [['p', 'i', '('], sshow f, [')', ' ', '=', ' '], sshow_nat n] :
    create_sp_representation (SimpProj sp);

check_for_var ::
  forall f v. (Sshow f, Sshow v) => (Term f v, Term f v) -> Eithera [[Char]] ();
check_for_var rule =
  catcha
    (checkB (not (isVar (fst rule)))
      [['t', 'h', 'e', ' ', 'l', 'h', 's', ' ', 'i', 's', ' ', 'a', ' ', 'v',
         'a', 'r', 'i', 'a', 'b', 'l', 'e']])
    (\ x ->
      errora
        (concata
           [sshow_term (fst rule),
             [' ', 'i', 's', ' ', 'a', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l',
               'e']]
          : x));

check_supteq ::
  forall f v.
    (Eq f, Sshow f, Eq v,
      Sshow v) => (f -> Nat) -> (Term f v, Term f v) -> Eithera [[Char]] ();
check_supteq p lr =
  catcha
    (checkB
      (lift2term p (fst lr) == lift2term p (snd lr) ||
        checkSupterm (lift2term p (fst lr)) (lift2term p (snd lr)))
      [['t', 'h', 'e', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', ' ',
         'r', 'h', 's', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 's',
         'u', 'b', 't', 'e', 'r', 'm', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ',
         'p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', ' ', 'l', 'h', 's']])
    (\ x ->
      errora
        (concata
           [['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
              'e', 'n', 't', ' ', 'r', 'u', 'l', 'e', ' '],
             sshow_term (fst lr), [' ', '-', '>', ' '], sshow_term (snd lr)]
          : concata
              [['i', '.', 'e', '.', ',', ' '],
                sshow_term (lift2term p (fst lr)), [' ', '|', '>', '=', ' '],
                sshow_term (lift2term p (snd lr))]
              : x));

check_projection ::
  forall a b.
    (Eq a, Sshow a,
      Sshow b) => [a] ->
                    (a -> Nat) -> (Term a b, Term a b) -> Eithera [[Char]] ();
check_projection ds p (Fun f ss, Fun g ts) =
  bind (bind (bind (bind (bind (checkB (less_nat Zero_nat (size_list ss))
                                 [['l', 'h', 's', ' ', 'i', 's', ' ', 'a', ' ',
                                    'c', 'o', 'n', 's', 't', 'a', 'n', 't']])
                           (\ x ->
                             checkB (not (member f ds))
                               [concata
                                  [sshow f,
                                    [' ', 'i', 's', ' ', 'd', 'e', 'f', 'i',
                                      'n', 'e', 'd', ' ', 'i', 'n', ' ',
                                      'R']]]))
                     (\ x ->
                       checkB (less_nat (p f) (size_list ss))
                         [concata
                            [sshow_nat (p f),
                              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n',
                                ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n', 't',
                                ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n',
                                ' ', 'o', 'f', ' '],
                              sshow_term (Fun f ss)]]))
               (\ x ->
                 checkB (less_nat Zero_nat (size_list ts))
                   [['r', 'h', 's', ' ', 'i', 's', ' ', 'a', ' ', 'c', 'o', 'n',
                      's', 't', 'a', 'n', 't']]))
         (\ x ->
           checkB (not (member g ds))
             [concata
                [sshow g,
                  [' ', 'i', 's', ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd', ' ',
                    'i', 'n', ' ', 'R']]]))
    (\ x ->
      checkB (less_nat (p g) (size_list ts))
        [concata
           [sshow_nat (p g),
             [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n', ' ', 'a', 'r',
               'g', 'u', 'm', 'e', 'n', 't', ' ', 'p', 'o', 's', 'i', 't', 'i',
               'o', 'n', ' ', 'o', 'f', ' '],
             sshow_term (Fun g ts)]]);
check_projection ds p (Var v, t) =
  errora
    [['c', 'a', 'n', 'n', 'o', 't', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't', ' ',
       'a', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e']];
check_projection ds p (s, Var v) =
  errora
    [['c', 'a', 'n', 'n', 'o', 't', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't', ' ',
       'a', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e']];

check_supt ::
  forall f v.
    (Eq f, Sshow f, Eq v,
      Sshow v) => (f -> Nat) -> (Term f v, Term f v) -> Eithera [[Char]] ();
check_supt p lr =
  catcha
    (checkB (checkSupterm (lift2term p (fst lr)) (lift2term p (snd lr)))
      [['t', 'h', 'e', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', ' ',
         'r', 'h', 's', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p',
         'r', 'o', 'p', 'e', 'r', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', ' ',
         'o', 'f', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't',
         'e', 'd', ' ', 'l', 'h', 's']])
    (\ x ->
      errora
        (concata
           [['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
              'e', 'n', 't', ' ', 'r', 'u', 'l', 'e', ' '],
             sshow_term (fst lr), [' ', '-', '>', ' '], sshow_term (snd lr)]
          : concata
              [['i', '.', 'e', '.', ',', ' '],
                sshow_term (lift2term p (fst lr)), [' ', '|', '>', ' '],
                sshow_term (lift2term p (snd lr))]
              : x));

check_spsc_proc ::
  forall a b.
    (Eq a, Sshow a, Eq b,
      Sshow b) => (a -> Nat) ->
                    [[Char]] ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          [(Term a b, Term a b)] -> Eithera [[Char]] ();
check_spsc_proc proj rep p rP r =
  let {
    ds = computeDSymbs r;
    a = computeSplit (\ ra -> isValue (check_supt proj ra)) p;
    (sp, nSp) = a;
  } in catcha
         (bind (bind (bind (catcha
                             (catcha (checkAllMI (check_projection ds proj) p)
                               (\ x -> errora (snd x)))
                             (\ x ->
                               errora
                                 (['t', 'h', 'e', ' ', 's', 'u', 'p', 'p', 'l',
                                    'i', 'e', 'd', ' ', 'p', 'r', 'o', 'j', 'e',
                                    'c', 't', 'i', 'o', 'n', ' ', 'i', 's', ' ',
                                    'm', 'a', 'l', 'f', 'o', 'r', 'm', 'e', 'd']
                                   : x)))
                       (\ x ->
                         catcha
                           (catcha (checkAllMI (check_supteq proj) nSp)
                             (\ xa -> errora (snd xa)))
                           (\ xa ->
                             errora
                               (['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                  'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n',
                                  't', 'i', 'n', 'g', ' ', 'D', 'P', 's']
                                 : xa))))
                 (\ x ->
                   catcha
                     (catcha (checkAllMI check_for_var r)
                       (\ xa -> errora (snd xa)))
                     (\ xa ->
                       errora
                         (['o', 'n', 'e', ' ', 'o', 'f', ' ', 't', 'h', 'e',
                            ' ', 'l', 'h', 's', 's', ' ', 'o', 'f', ' ', 'R',
                            ' ', 'i', 's', ' ', 'a', ' ', 'v', 'a', 'r', 'i',
                            'a', 'b', 'l', 'e']
                           : xa))))
           (\ x ->
             catcha (checkSubset nSp rP)
               (\ xa ->
                 errora
                   (case check_supt proj xa of {
                     Lefta err -> err;
                     Righta ok ->
                       [['i', 'n', 't', 'e', 'r', 'n', 'a', 'l', ' ', 'e', 'r',
                          'r', 'o', 'r'],
                         ['p', 'l', 'e', 'a', 's', 'e', ' ', 'c', 'o', 'n', 't',
                           'a', 'c', 't', ' ', 'o', 'n', 'e', ' ', 'o', 'f',
                           ' ', 't', 'h', 'e', ' ', 'C', 'e', 'T', 'A', ' ',
                           'm', 'e', 'm', 'b', 'e', 'r', 's'],
                         ['h', 't', 't', 'p', ':', '/', '/', 'c', 'l', '-', 'i',
                           'n', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'k', '.',
                           'u', 'i', 'b', 'k', '.', 'a', 'c', '.', 'a', 't',
                           '/', 's', 'o', 'f', 't', 'w', 'a', 'r', 'e', '/',
                           'c', 'e', 't', 'a'],
                         ['c', 'e', 't', 'a', '@', 'i', 'n', 'f', 'o', 'r', 'm',
                           'a', 't', 'i', 'k', '.', 'u', 'i', 'b', 'k', '.',
                           'a', 'c', '.', 'a', 't']];
                   }))))
         (\ x ->
           errora
             (['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
                'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'u', 'b', 't', 'e', 'r',
                'm', ' ', 'c', 'r', 'i', 't', 'e', 'r', 'i', 'o', 'n', ' ', 'p',
                'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h',
                ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n',
                'g', ' ', 's', 'i', 'm', 'p', 'l', 'e', ' ', 'p', 'r', 'o', 'j',
                'e', 'c', 't', 'i', 'o', 'n']
               : append rep ([] : x)));

create_sp :: forall a. (Eq a) => SpL a -> a -> Nat;
create_sp (SimpProj sp) f =
  (case lookupa f sp of {
    Nothing -> Zero_nat;
    Just n -> n;
  });

create_spsc_proc ::
  forall a b.
    (Eq a, Sshow a, Eq b,
      Sshow b) => SpL a ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] -> Eithera [[Char]] ();
create_spsc_proc sp p rP r =
  let {
    pia = create_sp sp;
  } in check_spsc_proc pia (create_sp_representation sp) p rP r;

}
