(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matrix imports  Interpretation
begin




(* matrices *)

types 'a vec = "'a list"

locale vecorder = 
  fixes carrier_s :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix ">" 50)
   and  carrier_ns :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<ge>" 50)
  assumes SN: "SN {(x,y) . (x :: 'a) > y}"
   and  compat: "\<lbrakk>s \<ge> (t :: 'a); t > u\<rbrakk> \<Longrightarrow> s > u"
   and  geq_trans: "\<lbrakk>s \<ge> (t :: 'a); t \<ge> u\<rbrakk> \<Longrightarrow> s \<ge> u"
begin
fun wf_vec :: "nat \<Rightarrow> 'x vec \<Rightarrow> bool"
  where "wf_vec n x = (length x = n)"


fun vec_ns :: "'a vec \<Rightarrow> nat \<Rightarrow> 'a vec \<Rightarrow> bool" ("(_ \<succeq> _ _)" [51,51,51] 50) 
  where "v \<succeq>n w = (wf_vec n v \<and> wf_vec n w \<and> (\<forall> i < length v. v ! i \<ge> w ! i))"
fun vec_s :: "'a vec \<Rightarrow> nat \<Rightarrow> 'a vec \<Rightarrow> bool" ("(_ \<succ> _ _)" [51,51,51] 50)
  where "v \<succ>n w = (v \<succeq>n w \<and> (\<exists> i < length v. v ! i > w ! i))"

lemma vec_ns_trans : assumes "u \<succeq>n v" and "v \<succeq>n w" 
  shows "u \<succeq>n w"
  using assms
proof (simp, clarify)
  fix i
  assume "i < length u"
  with assms have "u ! i \<ge> v ! i" and "v ! i \<ge> w ! i" by auto
  with geq_trans show "u ! i \<ge> w ! i" by blast
qed  

lemma vec_compat : assumes "u \<succeq>n v" and "v \<succ>n w" shows "u \<succ>n w" 
using assms
proof (simp only: vec_s.simps, simp only: vec_ns_trans[where u = u and v = v and w = w], clarify)
  fix i 
  assume "i < length v" and s: "v ! i > w ! i"
  with assms have i: "i < length u" and "u ! i \<ge> v ! i" by auto
  with s compat have "u ! i > w ! i" by blast
  with i show "\<exists> i < length u. u ! i > w ! i" by auto
qed  

lemma vec_sn : "SN {(v,w) . v \<succ>n w}"
unfolding SN_ars_def SN_elt_def 
proof clarify
  fix e f 
  assume ass: "\<forall> i. (f i, f (Suc i)) \<in> {(v,w). v \<succ>n w}"
  hence ns_vec: "\<And> i. (f i \<succeq>n f (Suc i))" by auto
  hence len_n: "\<And> i. length (f i) = n" by auto
  with ns_vec have ns: "\<And> i j. j < n \<Longrightarrow> f i ! j \<ge> f (Suc i) ! j" by auto 
  from ass len_n have s : "\<And> i. \<exists> j < n. f i ! j > f (Suc i) ! j" by auto

  show "False"
  proof (cases "\<exists> j < n. \<forall> m . \<exists> i . i \<ge> m \<and> f i ! j > f (Suc i) ! j")
    case "False"
    with False have "\<forall> j < n. \<exists> m . \<forall> i . i \<ge> m \<longrightarrow> \<not> (f i ! j > f (Suc i) ! j)" by blast
    hence "\<exists> m. \<forall> j < n. \<forall> i. i \<ge> m \<longrightarrow> \<not> (f i ! j > f (Suc i) ! j)"
    proof (induct n, simp)
      case (Suc nn)
      from Suc obtain m1 where res1: "\<forall> j < nn. \<forall> i \<ge> m1 . \<not> (f i ! j > f (Suc i) ! j)" by auto
      from Suc obtain m2 where res2: "\<forall> i \<ge> m2. \<not> (f i ! nn > f (Suc i) ! nn)" by auto
      let ?m = "max m1 m2"
      have "\<forall> j < Suc nn. \<forall> i \<ge> ?m. \<not> (f i ! j > f (Suc i) ! j)" 
      proof (rule allI)
	fix j
	show "j < Suc nn \<longrightarrow> (\<forall> i \<ge> ?m . \<not> (f i ! j > f (Suc i) ! j))"
	  using res1 res2 by (cases "j = nn", auto)
      qed
      thus ?case ..
    qed
    with s show False by blast      
  next
    case "True"
    from this obtain j where jlen: "j < n" and jinf: "\<forall> m. \<exists> i \<ge> m. f i ! j > f (Suc i) ! j" by auto
    obtain g where gdef: "g = (\<lambda> i. f i ! j)" by auto
    let ?S = "{(x,y). (x :: 'a) > y}"
    let ?NS = "{(x,y). (x :: 'a) \<ge> y}"
    from SN have SNstart: "SN ?S (g 0)" unfolding SN_ars_def by simp
    from ns jlen gdef have steps: "\<forall> i. (g i, g (Suc i)) \<in> ?NS \<union> ?S" by simp
    from compat have "?S O ?NS \<subseteq> ?S" by auto
    from steps this SNstart have "\<exists> m. \<forall> i \<ge> m. (g i, g (Suc i)) \<in> ?NS - ?S" by (rule non_strict_ending)
    with gdef jinf show "False" by blast
  qed
qed
end



interpretation nat_vec : vecorder "op > :: nat \<Rightarrow> nat \<Rightarrow> bool" "op \<ge>"
  by (unfold_locales, auto simp: nat_wfcarrier.SN)

(* the following interpretation should be made for every vecorder!
  vecorder \<subseteq> wforder [using vecS vecNS] *)
interpretation nat_vec_varrier : wfcarrier  "vecorder.vec_s (op > :: nat \<Rightarrow> nat \<Rightarrow> bool) (op \<ge>)" "vecorder.vec_ns (op \<ge>)"
  by (unfold_locales, rule nat_vec.vec_sn, simp only: nat_vec.vec_compat)

types 'a mat = "'a list list"

locale vecops = vecorder + 
  fixes plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
  and   zero :: "'a"
  and   add  :: "'a \<Rightarrow> 'd \<Rightarrow> 'a"
  and   mul :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
  assumes plus_mono : "\<lbrakk>x1 \<ge> x2; y1 \<ge> y2\<rbrakk> \<Longrightarrow> plus x1 y1 \<ge> plus x2 y2"
  and   mul_mono : "\<lbrakk>x1 \<ge> x2; y1 \<ge> y2\<rbrakk> \<Longrightarrow> mul x1 y1 \<ge> mul x2 y2"
  and   geq_refl : "(x :: 'a) \<ge> x"
  and   add_mono : "x \<ge> y \<Longrightarrow> add x d \<ge> add y d"
begin


fun wf_mat_rc :: "nat \<Rightarrow> nat \<Rightarrow> 'x list list \<Rightarrow> bool" where
  "wf_mat_rc nr nc m = (length m = nr \<and> (\<forall> i < nr. wf_vec nc (m ! i)))"

abbreviation wf_mat :: "nat \<Rightarrow> 'x mat \<Rightarrow> bool"
  where "wf_mat n \<equiv> wf_mat_rc n n"

fun vec_plus :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a vec" where 
  "vec_plus v w = map (\<lambda> xy. plus (fst xy) (snd xy)) (zip v w)"

fun vec_zero :: "nat \<Rightarrow> 'a vec" where 
  "vec_zero n = replicate n zero"

lemma wf_vec_zero[simp] : "wf_vec n (vec_zero n)" 
  by simp

lemma vec_plus_mono : assumes "v1 \<succeq>n v2" and "w1 \<succeq>n w2" shows "vec_plus v1 w1 \<succeq>n (vec_plus v2 w2)"
using assms
proof (simp, clarify)
  fix i 
  assume "i < length v1"
  with assms have "v1 ! i \<ge> v2 ! i" and "w1 ! i \<ge> w2 ! i" by auto
  with plus_mono show "plus (v1 ! i) (w1 ! i) \<ge> plus (v2 ! i) (w2 ! i)" by blast
qed

lemma vec_geq_refl : assumes "wf_vec n x" shows "x \<succeq>n x"
  using assms geq_refl by simp

lemma vec_plus_mono_l : assumes "v1 \<succeq>n v2" and "wf_vec n w" shows "vec_plus v1 w \<succeq>n (vec_plus v2 w)"
proof -
  from assms have "w \<succeq>n w" by (simp only: vec_geq_refl)
  with assms show ?thesis using vec_plus_mono by auto
qed


fun vec_add :: "'a vec \<Rightarrow> 'd vec \<Rightarrow> 'a vec" where 
  "vec_add v w = map (\<lambda> xy. add (fst xy) (snd xy)) (zip v w)"

lemma vec_add_mono : assumes "v1 \<succeq>n v2" and "wf_vec n w" shows "vec_add v1 w \<succeq>n (vec_add v2 w)"
using assms
proof (simp, clarify)
  fix i 
  assume "i < length w"
  with assms have "v1 ! i \<ge> v2 ! i" by simp
  with add_mono show "add (v1 ! i) (w ! i) \<ge> add (v2 ! i) (w ! i)" by blast
qed

fun scalar_prod :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a" where
  "scalar_prod v w = foldr (\<lambda> xy s. plus (mul (fst xy) (snd xy)) s) (zip v w) zero"

lemma scalar_prod_mono : assumes "v1 \<succeq>n v2" and "w1 \<succeq>n w2" shows "scalar_prod v1 w1 \<ge> scalar_prod v2 w2"
using assms
proof (induct n arbitrary: v1 v2 w1 w2, simp add: geq_refl)
  case (Suc nn)
  thus ?case proof (cases v1, simp, cases v2, simp, cases w1, simp, cases w2, simp)
    fix x1 vv1 x2 vv2 y1 ww1 y2 ww2
    assume eq1: "v1 = x1 # vv1" and eq2: "v2 = x2 # vv2" 
    with Suc have x: "x1 \<ge> x2" and vv: "vv1 \<succeq>nn vv2" by auto
    assume eq3: "w1 = y1 # ww1" and eq4: "w2 = y2 # ww2" 
    with Suc have y: "y1 \<ge> y2" and ww: "ww1 \<succeq>nn ww2" by auto
    from vv ww Suc have sc: "scalar_prod vv1 ww1 \<ge> scalar_prod vv2 ww2" by auto
    from x y mul_mono have mul: "mul x1 y1 \<ge> mul x2 y2" by blast
    from eq1 eq2 eq3 eq4 have "scalar_prod v1 w1 = plus (mul x1 y1) (scalar_prod vv1 ww1) \<and> 
      scalar_prod v2 w2 = plus (mul x2 y2) (scalar_prod vv2 ww2)" by auto
    with sc mul plus_mono[where ?x1.0 = "mul x1 y1"] show "scalar_prod v1 w1 \<ge> scalar_prod v2 w2" by auto
  qed
qed


lemma scalar_prod_mono_l : assumes "wf_vec n w" and "v1 \<succeq>n v2" shows "scalar_prod v1 w \<ge> scalar_prod v2 w"
proof (rule scalar_prod_mono)
  from assms vec_geq_refl show "w \<succeq>n w" by auto
next
  from assms show "v1 \<succeq>n v2" by simp
qed

fun mat_vec_mul :: "'a mat \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
  where "mat_vec_mul m v = map (scalar_prod v) m" 

lemma mat_vec_mul_mono : assumes "wf_mat_rc nr nc m" and "v \<succeq>nc w" shows "mat_vec_mul m v \<succeq>nr (mat_vec_mul m w)" 
  using assms by (auto simp: scalar_prod_mono_l simp del: scalar_prod.simps)

end


interpretation cint_vec_ops : vecops "op > :: nat \<Rightarrow> nat \<Rightarrow> bool" "op \<ge>" "op +" 0 "add" "op *"
 by (unfold_locales, auto simp: add_monotone Ring_and_Field.pordered_semiring_class.mult_mono)

interpretation nat_vec_ops : vecops "op > :: nat \<Rightarrow> nat \<Rightarrow> bool" "op \<ge>" "op +" 0 "op +" "op *"
  by (unfold_locales, auto simp: Ring_and_Field.pordered_semiring_class.mult_mono)


interpretation cint_mat_poly_inter : polyinter "vecorder.vec_s (op > :: nat \<Rightarrow> nat \<Rightarrow> bool) (op \<ge>)" "vecorder.vec_ns (op \<ge>)" "vecops.vec_plus (op +)" "vecops.vec_zero 0" "vecops.vec_add add" "vecops.mat_vec_mul (op +) 0 (op * )" "vecorder.wf_vec" "vecops.wf_mat" "vecorder.wf_vec"
  by (unfold_locales, 
      simp only: cint_vec_ops.vec_add_mono, 
      simp only: cint_vec_ops.vec_plus_mono,
      simp only: cint_vec_ops.mat_vec_mul_mono,
      simp only: cint_vec_ops.vec_geq_refl)

interpretation nat_mat_poly_inter : polyinter "vecorder.vec_s (op > :: nat \<Rightarrow> nat \<Rightarrow> bool) (op \<ge>)" "vecorder.vec_ns (op \<ge>)" "vecops.vec_plus (op +)" "vecops.vec_zero 0" "vecops.vec_plus (op +)" "vecops.mat_vec_mul (op +) 0 (op * )" "vecorder.wf_vec" "vecops.wf_mat" "vecorder.wf_vec"
  by (unfold_locales, 
      simp only: nat_vec_ops.vec_plus_mono_l,
      simp only: nat_vec_ops.vec_plus_mono,
      simp only: nat_vec_ops.mat_vec_mul_mono,
      simp only: nat_vec_ops.vec_geq_refl)

locale matops = vecops +
  fixes add_d :: "'d \<Rightarrow> 'd \<Rightarrow> 'd" 
  and   zero_d :: "'d"           
  and   mul_d :: "'a \<Rightarrow> 'd \<Rightarrow> 'd" 
  and   one :: "'a" 
  and   max_zero :: "'d \<Rightarrow> 'd" 
  and   geq_d :: "'d \<Rightarrow> 'd \<Rightarrow> bool" 
  and   gt_d :: "'d \<Rightarrow> 'd \<Rightarrow> bool"
  assumes plus_comm: "plus (x :: 'a) y = plus y x"
  and     mul_comm: "mul (x :: 'a) y = mul y x"
  and     plus_assoc[simp]: "plus (plus x y) z = plus x (plus y z)"
  and     mul_distrib_r[simp]: "mul x (plus y z) = plus (mul x y) (mul x z)"
  and     mul_distrib_l[simp]: "mul (plus x y) z = plus (mul x z) (mul y z)"
  and     mul_zero_r[simp]: "mul x zero = zero"
  and     mul_zero_l[simp]: "mul zero x = zero"
  and     plus_zero_r[simp]: "plus x zero = x"
  and     plus_zero_l[simp]: "plus zero x = x"

interpretation nat_mat_ops : matops "op > :: nat \<Rightarrow> nat \<Rightarrow> bool" "op \<ge>" "op +" 0 "op +" "op *" "op +" 0 "op *" 1 id "op \<ge>" "op >"
  by (unfold_locales, auto simp: nat_distrib )

interpretation cint_mat_ops : matops "op > :: nat \<Rightarrow> nat \<Rightarrow> bool" "op \<ge>" "op +" 0 "add" "op *" addI "Pos 0" mul 1 cint_max_zero geq_cint gt_cint
  by (unfold_locales, auto simp: add_monotone nat_distrib)

context matops
begin

lemma vec_plus_comm: assumes "wf_vec n u" and "wf_vec n v" 
  (* these wfVecs are not necessary, but every other op uses wfVecs
     => added these for consistent usage *)
  shows "vec_plus u v = vec_plus v u" 
proof (induct u arbitrary: v, simp)
  case (Cons x uu)
  thus ?case
  proof (cases v, simp)
    case (Cons y vv)
    from prems plus_comm[where x = x and y = y] show ?thesis by auto
  qed
qed

lemma vec_plus_assoc[simp]: assumes "wf_vec n u" and "wf_vec n v" and "wf_vec n w"
  shows "vec_plus (vec_plus u v) w = vec_plus u (vec_plus v w)"
using assms
proof (induct n arbitrary: u v w, simp)
  case (Suc nn)
  thus ?case 
  proof (cases u, simp, cases v, simp, cases w, simp)
    fix x uu y vv z ww
    assume "u = x # uu" and "v = y # vv" and "w = z # ww"
    with Suc show "vec_plus (vec_plus u v) w = vec_plus u (vec_plus v w)" by simp
  qed
qed

lemma vec_plus_zero_r[simp]: assumes "wf_vec n u" shows "vec_plus u (vec_zero n) = u"
using assms
proof (induct n arbitrary: u, simp)
  case (Suc nn) thus ?case by (cases u, auto)
qed

lemma vec_plus_zero_l[simp]: assumes "wf_vec n u" shows "vec_plus (vec_zero n) u = u"
using assms
proof (induct n arbitrary: u, simp)
  case (Suc nn) thus ?case by (cases u, auto)
qed

lemma scalar_prod_zero_l[simp]: assumes "wf_vec n u" shows "scalar_prod (vec_zero n) u = zero"
proof (induct n arbitrary: u, simp)
  case (Suc nn)
  thus ?case by (cases u, auto)
qed

lemma scalar_prod_zero_r[simp]: assumes "wf_vec n u" shows "scalar_prod u (vec_zero n) = zero"
proof (induct n arbitrary: u, simp)
  case (Suc nn)
  thus ?case by (cases u, auto)
qed

lemma vec_mul_zero[simp]: assumes "wf_mat_rc nr nc m" shows "mat_vec_mul m (vec_zero nc) = vec_zero nr"
using assms 
proof (induct nr arbitrary: m, simp)
  case (Suc nrr) thus ?case 
  proof (cases m, simp)
    case (Cons v mm)
    with Suc have one: "scalar_prod (vec_zero nc) v = zero" by (auto simp del: vec_zero.simps scalar_prod.simps)
    from Cons Suc have "mat_vec_mul mm (vec_zero nc) = vec_zero nrr" by force
    with one Suc Cons show ?thesis by auto
  qed
qed

lemma scalar_plus_distrib_l[simp]: assumes "wf_vec n u" and "wf_vec n v" and "wf_vec n w" 
  shows "scalar_prod (vec_plus u v) w = plus (scalar_prod u w) (scalar_prod v w)"
using assms
proof (induct n arbitrary: u v w, simp)
  case (Suc nn)
  thus ?case proof (cases u, simp, cases v, simp, cases w, simp)
    fix x uu y vv z ww
    assume eq1: "u = x # uu" and eq2: "v = y # vv" and eq3: "w = z # ww"
    from eq1 eq2 eq3 have "scalar_prod (vec_plus u v) w = plus ( plus (mul x z) (mul y z)) (scalar_prod (vec_plus uu vv) ww)" by simp
    also have "\<dots> = plus ( plus (mul x z) (mul y z)) (plus (scalar_prod uu ww) (scalar_prod vv ww))" using Suc eq1 eq2 eq3 by simp
    also have "\<dots> = plus (plus (mul x z) (scalar_prod uu ww))  (plus (mul y z) (scalar_prod vv ww))" using plus_comm plus_assoc by arith
    also have "\<dots> = plus (scalar_prod u w) (scalar_prod v w)" using Suc eq1 eq2 eq3 by simp
    finally show ?thesis .
  qed
qed

lemma scalar_plus_distrib_r[simp]: assumes "wf_vec n u" and "wf_vec n v" and "wf_vec n w" 
  shows "scalar_prod u (vec_plus v w) = plus (scalar_prod u v) (scalar_prod u w)"
using assms
proof (induct n arbitrary: u v w, simp)
  case (Suc nn)
  thus ?case proof (cases u, simp, cases v, simp, cases w, simp)
    fix x uu y vv z ww
    assume eq1: "u = x # uu" and eq2: "v = y # vv" and eq3: "w = z # ww"
    from eq1 eq2 eq3 have "scalar_prod u (vec_plus v w) = plus ( plus (mul x y) (mul x z)) (scalar_prod uu (vec_plus vv ww))" by simp
    also have "\<dots> = plus ( plus (mul x y) (mul x z)) (plus (scalar_prod uu vv) (scalar_prod uu ww))" using Suc eq1 eq2 eq3 by simp
    also have "\<dots> = plus (plus (mul x y) (scalar_prod uu vv))  (plus (mul x z) (scalar_prod uu ww))" using plus_comm plus_assoc by arith
    also have "\<dots> = plus (scalar_prod u v) (scalar_prod u w)" using Suc eq1 eq2 eq3 by simp
    finally show ?thesis .
  qed
qed

declare scalar_prod.simps[simp del]
  

lemma mat_vec_mul_distrib[simp]: assumes "wf_mat_rc nr nc m" and "wf_vec nc u" and "wf_vec nc v" 
  shows "mat_vec_mul m (vec_plus u v) = vec_plus (mat_vec_mul m u) (mat_vec_mul m v)"
using assms
proof (induct nr arbitrary: m, simp)
  case (Suc nrr) thus ?case proof (cases m, simp)
    case (Cons w mm)
    from Cons Suc have one: "mat_vec_mul mm (vec_plus u v) = vec_plus (mat_vec_mul mm u) (mat_vec_mul mm v)" by force
    from Cons Suc have two: "scalar_prod (vec_plus u v) w = plus (scalar_prod u w) (scalar_prod v w)" by (auto simp del: vec_plus.simps)
    from Cons Suc one two show "mat_vec_mul m (vec_plus u v) = vec_plus (mat_vec_mul m u) (mat_vec_mul m v)" by auto
  qed
qed


fun mat_zero :: "nat \<Rightarrow> nat \<Rightarrow> 'a mat"
  where "mat_zero nr nc = replicate nr (vec_zero nc)"

lemma wf_mat_zero : "wf_mat_rc nr nc (mat_zero nr nc)"
  by simp


lemma mat_vec_mul_zero[simp]: assumes "wf_vec nc u" shows "mat_vec_mul (mat_zero nr nc) u = vec_zero nr"
using assms
proof (induct nr, simp)
  case (Suc nrr)
  hence one: "mat_vec_mul (mat_zero nrr nc) u = vec_zero nrr" by auto
  from Suc have "scalar_prod u (vec_zero nc) = zero" by (simp only: scalar_prod_zero_r)
  with one Suc show ?case by simp
qed

fun mat_plus :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
  where "mat_plus m1 m2 = map (\<lambda> uv. vec_plus (fst uv) (snd uv)) (zip m1 m2)"

lemma mat_vec_mul_plus_distrib : assumes wf1: "wf_mat_rc nr nc m1" and wf2: "wf_mat_rc nr nc m2" and wf3: "wf_vec nc w"
  shows "mat_vec_mul (mat_plus m1 m2) w = vec_plus (mat_vec_mul m1 w) (mat_vec_mul m2 w)"
using wf1 wf2
proof (induct nr arbitrary: m1 m2, simp)
  case (Suc nrr) thus ?case proof (cases m1, simp, cases m2, simp)
    fix u m11 v m22
    assume eq1: "m1 = u # m11" and eq2: "m2 = v # m22" 
    from eq1 eq2 Suc have "wf_mat_rc nrr nc m11" and "wf_mat_rc nrr nc m22" by auto
    with Suc have one: "mat_vec_mul (mat_plus m11 m22) w = vec_plus (mat_vec_mul m11 w) (mat_vec_mul m22 w)" by auto
    from eq1 eq2 Suc have "wf_vec nc u" and "wf_vec nc v" by auto
    with wf3 Suc have two: "scalar_prod w (vec_plus u v) = plus (scalar_prod w u) (scalar_prod w v)" by (auto simp del: vec_plus.simps)
    from eq1 eq2 Suc one two show "mat_vec_mul (mat_plus m1 m2) w = vec_plus (mat_vec_mul m1 w) (mat_vec_mul m2 w)" by auto
  qed
qed
(*
function transpose :: "'a mat \<Rightarrow> 'a mat"
  where "transpose [] = []"
     |  "transpose ([] # m) = []"
     |  "transpose ((x # u) # m) = (let (c,mm) = foldr (\<lambda> v (d,mmm). (hd v # d, tl v # mmm)) m ([],[]) in ((x # c) # transpose mm))"
by (pat_completeness, auto)

termination transpose oops
*)


fun transpose :: "'a mat \<Rightarrow> 'a mat"
  where "transpose [] = []"
     |  "transpose ([] # m) = [[]]"
     |  "transpose ((x # u) # m) = (
           let mm = transpose m in
           case mm of [] \<Rightarrow> (x # []) # (map (\<lambda> y. y # []) u)
           | (Cons v mm')\<Rightarrow> (x # v) # (map (\<lambda> (y, w). y # w) (zip u mm')))"


lemma wf_mat_transp : 
 assumes "wf_mat_rc nr nc m" 
 shows "wf_mat_rc nc nr (transp m)"
 using assms 
oops (* "wfMatRC nr nc m = (length m = nr \<and> (\<forall> i < nr. wfVec nc (m ! i)))" *)


lemma wf_mat_transpose : assumes "wf_mat_rc nr nc m" shows "wf_mat_rc nr nc m" oops (* perhaps required *)

fun mat_mul :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat" 
 where "mat_mul m1 m2 = (let m3 = transpose m2 in map (mat_vec_mul m3) m1)"


lemma mat_mul_mat_vec_assoc[simp] : assumes wf1: "wf_mat_rc nc n m1" and wf2: "wf_mat_rc n nc m2" and wf3: "wf_vec nc w"
  shows "mat_vec_mul (mat_mul m1 m2) w = mat_vec_mul m1 (mat_vec_mul m2 w)" oops


end



(*
 TODO list from polyinter

  and     mulAadd: "mulA b (add a d) = add (mulA b a) (mulD b d)"
  and     maxZeroAddZero: "add (zero :: 'a) d \<ge> add zero (maxZero d)"
  and     maxZeroAdd: "add a (maxZero d) \<ge> add (a :: 'a) d"
  and     plusAdd:  "plus (add (a :: 'a) d) (add a' d') \<ge> add (plus a a') (addD d d')"
  and     addZeroD[simp]: "add a zeroD = a"
  and     mulAoneB[simp]: "mulA oneB a = a"
  and     geqGt: "a \<ge> (a' :: 'a) \<and> a' > a'' \<and> a'' \<ge> a''' \<Longrightarrow> a > a'''" 
  and     addAddD: "add (add (a :: 'a) (d :: 'd)) (d' :: 'd) \<ge> add a (addD d d')"
  and     posZeroD: "maxZero zeroD = zeroD"
  and     posMulD: "maxZero d = d \<Longrightarrow> maxZero (mulD b d) = mulD b d"
  and     posAddD: "\<lbrakk>maxZero d = d; maxZero d' = d'\<rbrakk> \<Longrightarrow> maxZero (addD d d') = addD d d'"
  and     posMaxZero: "maxZero (maxZero d) = maxZero d"
  and     posPlusAdd:  "\<lbrakk>maxZero d = d; maxZero d' = d'\<rbrakk> \<Longrightarrow> add (plus a a') (addD d d') = plus (add (a :: 'a) d) (add a' d')"
  and     posAddAddD: "maxZero d = d \<Longrightarrow> add a (addD d d') = add (add a d) d'"
  and     geqBMono: "geqB (b1 :: 'b) b2 \<Longrightarrow> (mulA b1 (a :: 'a)) \<ge> (mulA b2 a)"
  and     geqDMono: "geqD d d' \<Longrightarrow> add (a :: 'a) d \<ge> add a d'"
  and     gtDMono:  "gtD  d d' \<Longrightarrow> add (a :: 'a) d > add a d'"
  and     zeroLeast: "(a :: 'a) \<ge> zero"
  and     mulAMonoLeft: "geqB (b1 :: 'b) b2 \<Longrightarrow> (mulA b1 (a :: 'a) \<ge> mulA b2 a)"

*)

end 
