(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPairImpl imports Polynomial
begin

datatype 'f RedPair = NegPolo "('f \<times> (cint \<times> nat list)) list"


fun createRedPairProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
        sstring list monad"
where "createRedPairProc (NegPolo interpr) P rP R = (let \<alpha> = createPolyInter interpr in 
  checkRedPairProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>) 
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) interpr)))
           P rP R)" 

lemma redpair_proc_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rP. isOK(createRedPairProc redp P rP R))"
proof (cases redp)
  case (NegPolo interpr)
  let ?a = "createPolyInter interpr"
  have simpme: "createRedPairProc redp = checkRedPairProc (checkSPolo ?a) (checkNSPolo ?a) (''polynomial order'') 
      (createRepresentation ?a (remdups (map (\<lambda> (g,il) . (g, length (snd il))) interpr))) "
    by (rule ext, rule ext, rule ext, simp add: Let_def NegPolo)
  show ?thesis by (simp only: simpme, rule check_redpair_proc_sound)
qed

declare createRedPairProc.simps[simp del]

fun createMonoRedPairProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
        sstring list monad"
where "createMonoRedPairProc (NegPolo \<beta>) P rP R rR = (
  let \<alpha> = createPolyInter \<beta> in
      checkMonoRedPairProc 
           (checkInterMono \<beta>)
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>)  
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) \<beta>)))
           P rP R rR)"

lemma mono_redpair_proc_ci : 
  "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(createMonoRedPairProc redp P rP R rR))"
proof (cases redp)
  case (NegPolo interpr)
  let ?a = "createPolyInter interpr"
  have simpme: "createMonoRedPairProc redp = checkMonoRedPairProc (checkInterMono interpr) (checkSPolo ?a) (checkNSPolo ?a) (''polynomial order'') 
      (createRepresentation ?a (remdups (map (\<lambda> (g,il) . (g, length (snd il))) interpr))) "
    by (rule ext, rule ext, rule ext, rule ext, simp add: Let_def NegPolo)
  show ?thesis by (simp only: simpme, rule check_mono_redpair_proc_sound)
qed

declare createMonoRedPairProc.simps[simp del]


fun createRuleRemovalProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow>  ('f, 'v)trsL \<Rightarrow> 
        sstring list monad"
where "createRuleRemovalProc (NegPolo \<beta>) R rR = (
  let \<alpha> = createPolyInter \<beta> in
      checkRuleRemovalProc 
           (checkInterMono \<beta>)
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>)  
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) \<beta>)))
           R rR)"

lemma createRuleRemovalProcSound[simp] : fixes R rR
assumes check: "isOK(createRuleRemovalProc redp R rR)"
   shows "SN (rstep (set rR)) \<Longrightarrow> SN (rstep (set R))"
using assms by (cases redp, simp add: Let_def) 

fun createRedPairUrProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
       ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> sstring list monad"
where "createRedPairUrProc (NegPolo interpr) P rP R rm ur = (let \<alpha> = createPolyInter interpr in 
  checkRedPairUrProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>) 
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) interpr)))
           P rP R rm ur)" 

lemma redpair_ur_proc_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rP. isOK(createRedPairUrProc redp P rP R rm ur))"
proof (cases redp)
  case (NegPolo interpr)
  let ?a = "createPolyInter interpr"
  have simpme: "createRedPairUrProc redp  = checkRedPairUrProc (checkSPolo ?a) (checkNSPolo ?a) (''polynomial order'') 
      (createRepresentation ?a (remdups (map (\<lambda> (g,il) . (g, length (snd il))) interpr))) "
    by ((rule ext)+, simp add: Let_def NegPolo)
  show ?thesis by (simp only: simpme, rule check_redpair_ur_proc_sound)
qed

declare createRedPairUrProc.simps[simp del]

fun createRedPairUrAfProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> 
       ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> sstring list monad"
where "createRedPairUrAfProc (NegPolo l) P rP R rm ur = (let \<alpha> = createPolyInter l in
  checkRedPairUrAfProc 
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>) 
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) l)))
           P rP R rm ur (create_af l))" 


lemma redpair_ur_af_proc_ci : 
  "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rP. isOK(createRedPairUrAfProc redp P rP R rm ur))"
proof (cases redp)
  case (NegPolo interpr)
  let ?a = "createPolyInter interpr"
  have simpme: "\<And> P rP R rm. createRedPairUrAfProc redp P rP R rm ur = checkRedPairUrAfProc (checkSPolo ?a) (checkNSPolo ?a) (''polynomial order'') 
      (createRepresentation ?a (remdups (map (\<lambda> (g,il) . (g, length (snd il))) interpr))) P rP R rm ur (create_af interpr)"
    by (simp add: Let_def NegPolo)
  show ?thesis by (simp only: simpme, rule check_redpair_ur_af_proc_sound)
qed

declare createRedPairUrAfProc.simps[simp del]


fun createMonoRedPairUrProc :: "'f RedPair \<Rightarrow> ('f :: sshow ,'v :: sshow)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f, 'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow>
        sstring list monad"
where "createMonoRedPairUrProc (NegPolo \<beta>) P rP R rm rR ur = (
  let \<alpha> = createPolyInter \<beta> in (
   checkMonoRedPairUrProc 
           (checkInterMono \<beta>)
           (checkSPolo \<alpha>) 
           (checkNSPolo \<alpha>)  
           (''polynomial order'') 
           (createRepresentation \<alpha> (remdups (map (\<lambda> (g,il) . (g, length (snd il))) \<beta>)))
           P rP R rm rR ur))"

lemma mono_redpair_ur_proc_ci : 
  "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(createMonoRedPairUrProc redp P rP R rm rR ur))"
proof (cases redp)
  case (NegPolo interpr)
  let ?a = "createPolyInter interpr"
  have simpme: "\<And> P R rP rR rm. createMonoRedPairUrProc redp P rP R rm rR ur = checkMonoRedPairUrProc (checkInterMono interpr) (checkSPolo ?a) (checkNSPolo ?a) (''polynomial order'') 
      (createRepresentation ?a (remdups (map (\<lambda> (g,il) . (g, length (snd il))) interpr))) P rP R rm rR ur "
    by (simp add: Let_def NegPolo)
  show ?thesis by (simp only: simpme, rule check_mono_redpair_ur_proc_sound)
qed
  
end